/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Robert Bosch LLC.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Robert Bosch nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/

//\Author Philip Roan, Bosch LLC

#ifndef REMOTE_POWER_MANAGER_H_
#define REMOTE_POWER_MANAGER_H_


#include "remote_power_manager/RemotePowerPortStatus.h"
#include "remote_power_manager/GetPortStatus.h"
#include "remote_power_manager/SetAllRelay.h"
#include "remote_power_manager/SetPortRelay.h"
#include "remote_power_manager/RebootPort.h"

class RemotePowerManager
{
 public:
  /*!
   *
   * The constructor initializes a node to interface with a remote power manager. The NodeHandle is necessary so
   * variables can be initialized by parameters in the launch file.
   * \param nh   The node handle from the main loop
   */
  RemotePowerManager(const ros::NodeHandle& nh);
  //~RemotePowerManager();

  /*!
   * \brief Retrieve the status (on/off) of all power ports.
   *
   * Query the remote power manager over TCP to retrieve the status of all the
   * power ports controlled by the manager.
   * \param req   The ROS service request
   * \param res   The ROS service response which contains the power port status
   * \return      True if the status was correctly received
   */
  bool getPortStatus( remote_power_manager::GetPortStatus::Request  &req,
		      remote_power_manager::GetPortStatus::Response &res );
  /*!
   * \brief Set all power ports to the same value (on/off).
   * 
   * \param req   The ROS service request which contains the on/off value
   * \param res   The ROS service response
   * \return      True if the ports were correctly switched
   */
  bool setAllRelay( remote_power_manager::SetAllRelay::Request  &req,
		    remote_power_manager::SetAllRelay::Response &res );
  /*!
   * \brief Set the value (on/off) of a particular power port.
   * 
   * \param req   The ROS service request which contains the port and the on/off value
   * \param res   The ROS service response
   * \return      True if the port was correctly switched
   */
  bool setPortRelay( remote_power_manager::SetPortRelay::Request  &req,
		     remote_power_manager::SetPortRelay::Response &res );
  /*!
   * \brief Toggle the value (on/off) of a particular power port.
   * 
   * Rebooting a port will toggle the value to the opposite of the current value
   * and hold it there for the reboot time. After the reboot time has expired
   * the port will be reset to the original value. The reboot time is set in the
   * remote power manager from the web interface, and it cannot be set from this
   * package.
   * \param req   The ROS service request which contains the port
   * \param res   The ROS service response
   * \return      True if the reboot sequence was correctly initiated
   */
  bool rebootPort( remote_power_manager::RebootPort::Request  &req,
		   remote_power_manager::RebootPort::Response &res );

 private:
  int socket_;
  std::string host_name_;
  int host_port_;
  int number_of_power_ports_;

  /*!
   * \brief Parse out telnet commands in TCP data from the remote power manager.
   * 
   * Communication with the remote power manager is done through the telnet
   * protocal. These commands need to be acknowledged and responded to in order
   * for communication to proceed. This method parses telnet commands from a
   * character array and replies, only agreeing to drop the go ahead option.
   * \param raw_message   The character array including telnet control sequences.
   * \param length        The length of the input character array.
   * \param text_message  The new character array with no telnet sequences.
   * \return              The length of the character array of the new message.
   */
  int parseCommands(char* raw_message, int length, char* text_message);
  
  /*!
   * \brief Open a TCP connection to the host specified in the launch file.
   *
   * \return   True if the connection has been opened.
   */
  bool openConnection(void);

  /*!
   * \brief Close the TCP connection to the host specified in the launch file.
   *
   * \return   True if the connection has been closed.
   */  
  bool closeConnection(void);

  /*!
   * \brief Send a telnet command over the TCP connection specified in the launch file.
   *
   * The connection must be open before this function is called.
   * \param command   A character array giving the command to issue.
   * \param reply     A character array containing the response to the issued command.
   *
   * \return   True if the send operation is successful.
   */
  bool sendCommand(const char command[], char reply[]);
};
#endif
