/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Robert Bosch LLC.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Robert Bosch nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/
#ifndef VLRIMAGING_IMAGE_H_
#define VLRIMAGING_IMAGE_H_

#include <inttypes.h>
#include <algorithm>
#include <map>

#include "vlrTime.h"
#include "vlrImageTags.h"

namespace vlr {

class ImageBase
{
public:
		// Struct for (float) complex data compatible to Ipp32fc
	typedef struct {float re, im;} complex;

		// Struct for (double) complex data compatible to Ipp64fc
	typedef struct {double re, im;} dcomplex;

	typedef enum {BOMO_IGNORE=0, BOMO_CONSTANT, BOMO_REPLICATE, BOMO_REFLECT, BOMO_WRAP} borderMode_t;
	typedef enum {PAD_LEFT, PAD_SYM, PAD_RIGHT, PAD_TOP, PAD_BOTTOM, PAD_CUSTOM} padMode_t;

	typedef enum
		{
		CS_GRAY,    // Gray level
		CS_RGB,     // RGB
		CS_CMY,     // CMY
		CS_CMYK,    // CMYK
		CS_HSV,     // HSV
		CS_HLS,     // HLS
		CS_XYZ,     // XYZ
		CS_LUV,     // LUV
		CS_LAB,     // LAB
		CS_YCR,     // YCR
		CS_YCC,     // YCC
    CS_YUV,     // YUV
    CS_RGB_C    // RGB CHUNKY
		} colorSpace_t;

	typedef std::multimap<uint32_t, TagBase*>::const_iterator tagCIter;

protected:
	ImageBase(uint32_t width, uint32_t height,
			 uint32_t channels, uint32_t padded_width,
			 bool manage_buffer, colorSpace_t colorSpace) :
			 width_(width),
			 height_(height),
			 channels_(channels),
			 padded_width_(padded_width),
			 manage_buffer_(manage_buffer),
			 color_space_(colorSpace),
			 time_stamp_(currentTime())
	{
	num_elements_ = padded_width_*height_*channels_;
	}

public:
	virtual ~ImageBase() {}

public:
	inline uint32_t           width() const {return width_;}
	inline uint32_t           height() const {return height_;}
	inline uint32_t           channels() const {return channels_;}
	inline uint32_t           paddedWidth() const {return padded_width_;}
	inline bool			          manageBuffer() const {return manage_buffer_;}
	inline uint32_t           numElements() const {return num_elements_;}
	inline colorSpace_t	      colorSpace() const {return color_space_;}
	inline double	 	          timeStamp() const {return time_stamp_;}
  inline void               setTimeStamp(double time_stamp) {time_stamp_=time_stamp;}
  inline const std::string& name() const {return name_;}
  inline void               setName(const std::string& name)  {name_ = name;}
  inline const std::string& typeName() const {return type_name_;}

  inline TagPoint* addPoint(const float x, const float y, const std::string& label,
                                const float r, const float g, const float b, const float a) {
    TagPoint* pt = new TagPoint(x, y, label, r, g, b, a);
    tags_.insert(std::make_pair((uint32_t)tidPoint, pt));
    return pt;
  }

  inline TagPoint* addPoint(const float x, const float y, const std::string& label) {
    TagPoint* pt = new TagPoint(x, y, label);
    tags_.insert(std::make_pair((uint32_t)tidPoint, pt));
    return pt;
  }

  inline TagPoint* addPoint(const float x, const float y) {
    TagPoint* pt = new TagPoint(x, y);
    tags_.insert(std::make_pair((uint32_t)tidPoint, pt));
    return pt;
  }

	inline void points(std::pair<tagCIter, tagCIter>& range) {
		range=tags_.equal_range(tidPoint);
  }

  inline TagLine* addLine(const float x0, const float y0, const float x1, const float y1,
                              const std::string& label, const float r, const float g,
                              const float b, const float a) {
    TagLine* ln = new TagLine(x0, y0, x1, y1, label, r, g, b, a);
    tags_.insert(std::make_pair((uint32_t)tidLine, ln));
    return ln;
  }

  inline TagLine* addLine(const float x0, const float y0, const float x1, const float y1,
                              const std::string& label) {
    TagLine* ln = new TagLine(x0, y0, x1, y1, label);
    tags_.insert(std::make_pair((uint32_t)tidLine, ln));
    return ln;
  }

  inline TagLine* addLine(const float x0, const float y0, const float x1, const float y1) {
    TagLine* ln = new TagLine(x0, y0, x1, y1);
    tags_.insert(std::make_pair((uint32_t)tidLine, ln));
    return ln;
  }

  inline void lines(std::pair<tagCIter, tagCIter>& range) {
    range=tags_.equal_range(tidLine);
  }

  inline TagBase* addTag(TagBase* tag) {
    tags_.insert(std::make_pair((uint32_t)tag->id(), tag));
    return tag;
  }

  inline void clearTags() {
    for(std::multimap<uint32_t, TagBase*>::iterator it = tags_.begin(); it!=tags_.end(); ++it) {
      delete it->second;
    }
    tags_.clear();
  }

protected:
	uint32_t width_;
	uint32_t height_;
	uint32_t channels_;
	uint32_t padded_width_;
	uint32_t num_elements_;
	bool manage_buffer_;
	colorSpace_t color_space_;
	double time_stamp_;

	std::string name_;
	std::string description_;

	std::multimap<uint32_t, TagBase*> tags_;
  std::string type_name_;    // type name is used to make e.g. displays work with derived classes
};

template <class T> class Image : public ImageBase
{
public:
	Image(uint32_t width, uint32_t height, uint32_t channels, uint32_t padded_width,
           bool manage_buffer, colorSpace_t colorSpace);
  Image(uint32_t width, uint32_t height, uint32_t channels, colorSpace_t colorSpace);
  Image(uint32_t width, uint32_t height, uint32_t channels);
  Image(uint32_t width, uint32_t height);
  Image(const Image& img, bool manage_buffer=true, bool copyData=true, bool copyTags=true);

	Image& operator=(const Image& img);

	virtual ~Image();

		// inline functions
	inline uint32_t  elementSize() const {return element_size_;}
  inline T*        data()        const {return data_;}
  inline const T*  constData()  const {return (const T*)data_;}
	inline T*			   setData(T* new_data, bool manage_new_buffer=false) {
		T* tmp=data_;
		manage_buffer_= manage_new_buffer;
		data_= new_data;
		if(manage_buffer_ && tmp) {delete[] tmp; tmp=NULL;}

		return tmp;
		}

  inline bool sameDims(const Image& img) const {
    return (img.width() == width_ && img.height() == height_ &&
        img.channels() == channels_ && img.paddedWidth() == padded_width_);
    }

  template <class nT> operator Image<nT>() {
		Image<nT> res(width_, height_, channels_, padded_width_, true, color_space_);

    T* data_ptr = data_;
    nT* resdata = res.data();

    uint32_t align_gap = padded_width_ - width_;
    for (uint32_t c = 0; c < channels_; c++) {
      for (uint32_t y = 0; y < height_; y++) {
        for (uint32_t x = 0; x < width_; x++) {
          *resdata = nT(*data_ptr);
          resdata++;
          data_ptr++;
        }

        resdata += align_gap;
        data_ptr += align_gap;
      }
    }

    return res;
  }

	void bounds(T& lower_bound, T& upper_bound);
	void normalize(T new_min, T new_max);
	bool reformat(uint32_t new_width, uint32_t new_height, uint32_t new_channels,
										uint32_t new_padded_width, colorSpace_t new_color_space);

	Image<T> operator + (const Image<T>& img) const;

	inline T& operator [] (const uint32_t pos) const {return data_[pos];}
	inline T& operator () (const uint32_t x, const uint32_t y) const {return data_[y*padded_width_+x];}
	inline T& operator () (const uint32_t x, const uint32_t y, const uint32_t c) const
							 {return data_[c*padded_width_*height_+y*padded_width_+x];}
	Image<T> operator () (uint32_t x, uint32_t y, uint32_t roi_width, uint32_t roi_height) const;

private:
	void create();

private:
	uint32_t element_size_;
	T* data_;
	};

}	// namespace vlr

#endif // VLRIMAGING_IMAGE_H_
